import argparse
import csv
from itertools import count
from builtins import zip
import botocore
from boto3.session import Session
import plac
from ..saver import ObjectVersionSaver, FnameFactory


@plac.annotations(
    profile_name=("""Name of AWSCLI profile to use for credentials""",
                  "option"),
    aws_access_key_id=("Your AWS Access Key ID", "option"),
    aws_secret_access_key=("Your AWS Secret Access Key", "option"),
    bucket_name=("bucket name (default: %(default)s)", "positional"),
    csv_version_file=("name of CSV file with version_id", "positional", None,
                      argparse.FileType("r")),
    output_version_id_names=("Resulting file names shall use version_id "
                             "to become distinguished "
                             "(default is to use timestamp of file creation)",
                             "flag"),
    no_decompression=("Keeps the files as they come, do not decompress, "
                      "if they come compressed", "flag")
)
def main(bucket_name,
         csv_version_file,
         output_version_id_names=False,
         no_decompression=False,
         profile_name=None,
         aws_access_key_id=None,
         aws_secret_access_key=None):
    """Fetch S3 object versions as listed in a csv file

    Typical csv file (as by default produced by s3lsvers) is:

        m/y.xml;OrUr6XO8KSKEHbd8mQ.MloGcGlsh7Sir;191;2012-05-23T20:45:10.000Z;39
        m/y.xml;xhkVOy.dJfjSfUwse8tsieqjDicp0owq;192;2012-05-23T20:44:31.000Z;62
        m/y.xml;oKneK.N2wS8pW8.EmLqjldYlgcFwxN3V;193;2012-05-23T20:43:29.000Z;58

    for `s3getvers` only the first two columns are significant:
    :key_name: name of the object (not containing the bucket name itself)
    :version_id: string, identifying unique version.

    Typical use (assuming, above csv file is available under name verlist.csv)::

        $ %(prog)s yourbucketname verlist.csv

    What will create following files in current directory:

    * f.2012-05-23T20_45_10.xml
    * f.2012-05-23T20_44_31.xml
    * f.2012-05-23T20_43_29.xml

    Files are (by default) saved decompressed (even if gzipped on the bucket)
    """
    to_decompress = () if no_decompression else ("deflate", "gzip")

    session = Session(profile_name=profile_name,
                      aws_access_key_id=aws_access_key_id,
                      aws_secret_access_key=aws_secret_access_key)
    s3 = session.resource("s3")
    if output_version_id_names:
        fname_factory = FnameFactory().version_id
    else:
        fname_factory = FnameFactory().last_modified
    saver = ObjectVersionSaver(s3, fname_factory, to_decompress)
    try:
        recmsg = "{i:2d}: key_name: {key_name}, version_id: {version_id}"
        for i, row in zip(count(),
                          csv.reader(csv_version_file, delimiter=";")):
            key_name, version_id = row[:2]
            print(recmsg.format(**locals()))
            fname, key = saver.save(bucket_name, key_name, version_id)
            print("resulting file name: {fname}".format(fname=fname))
    except botocore.exceptions.ClientError as e:
        msg = ("Problem accessing bucket object.\n"
               "Possible causes:\n"
               "- missing/wrong credentials\n"
               "- missing bucket or object\n"
               "- no permission to access the object.")
        print(e)
        print(msg)
        return
    except KeyboardInterrupt:
        print("...terminated.")

    return


def placer():
    plac.call(main)

if __name__ == "__main__":
    plac.call(main)



import boto3
import argparse
import sys
import datetime
import itertools
import calendar
import plac
import os
import json
from builtins import zip
from csv import DictWriter
from configparser import SafeConfigParser
from ..utils import check_from_to

from importlib.resources import files
templ = files(__name__.rsplit(".", 2)[0]).joinpath('cli', 'chart.html').read_text(encoding="utf-8")


def load_alias(bucket_key):
    config = SafeConfigParser()
    rcfname = ".s3lsversrc"
    flocs = [os.curdir, os.path.expanduser("~"), "/etc/s3lsvers"]
    config.read([os.path.join(loc, rcfname) for loc in flocs])
    return config.get("aliases", bucket_key)


def get_json_record(dtime, size, age):
    tt = datetime.datetime.timetuple(dtime)
    epoch = calendar.timegm(tt) * 1000
    return [epoch, size, age]


def parse_bucket_key(bucket_key):
    if ("/" not in bucket_key):
        bucket_key = load_alias(bucket_key)
    assert "/" in bucket_key, "bucket_key must be in form 'bucket/key_name'"
    return bucket_key.split("/", 1)


def get_csvwriter(list_file):
    fields = ["key_name", "version_id", "size", "last_modified", "age"]
    if list_file:
        return DictWriter(list_file, fields,
                          delimiter=";", extrasaction="ignore", lineterminator="\n")
    else:
        return None


@plac.annotations(
    profile_name=("AWSCLI profile name", "option"),
    aws_access_key_id=("AWS Access Key ID", "option"),
    aws_secret_access_key=("AWS Secret Access Key", "option"),
    from_time=("start of version modification time range (default: oldest version)", "option", "from"),
    to_time=("end of version modification time range (default: now)", "option", "to"),
    list_file=("Name of output CSV file.", "option", None,
               argparse.FileType("w", encoding="UTF-8")),
    html_file=("Name of output HTML file.", "option", None,
               argparse.FileType("w")),
    version_id=("version-id to start after", "option"),
    bucket_key=("{bucket_name}/{key_name} for the key to list", "positional"))
def main(
        bucket_key,
        from_time=None,
        to_time=None,
        list_file=None,
        html_file=None,
        version_id=None,
        profile_name=None,
        aws_access_key_id=None,
        aws_secret_access_key=None):
    """List object versions stored on versioned S3 bucket, create CSV and/or HTML file.
    CSV file can be used e.g. by `s3getvers` command.
    HTML file allows showing feed size and update period in chart.

    Version can be limited by time range `from` - `to`.
    `version-id` allow starting from specific version (back to the past,
    excluding given version).

    Object key is defined either as {bucket_name}/{key_name} or as alias from .s3lsvers file.

    Times are expressed in RFC 3339 format using Zulu (UTC) timezone, possibly truncated.
    For truncated time strings, maximal time extent is used.

    Listing has records with structure:
      `{key_name};{version_id};{size};{last_modified};{age}`
        - key_name: name of the key (excluding bucket name).
        - version_id: unique identifier for given version on given bucket.
        - size: size of key object in bytes
        - last_modified: RFC 3339 formated object modification time
        - age: update interval [s] for given version

    Examples:

        Lists all versions of given `keyname` on `bucket`::

            $ s3lsvers bucketname/keyname

        Lists all versions in period betwen `from` and `to` time::

            $ s3lsvers -from 2010-01-01 -to 2011-07-19T12:00:00 bucket/key

        Lists all versions and writes them into csv file named `versions.csv`::

            $ s3lsvers -list-file versions.csv bucketname/keyname

        Lists all versions and write them into html chart file `chart.html`::

            $ s3lsvers -html-file chart.html bucketname/keyname

    Using bucket/key_name aliases in .s3lsvers file

        Aliases are specified in file .s3lsvers, which may be located in
        currect directory, home directory or /etc/s3lsvers"

        `.s3lsvers` example::

            #.s3lsversrc - definition of some preconfigured bucket/key values
            [DEFAULT]
            pl-base: pl-base.dp.tamtamresearch.com
            cz-base: cz-base.dp.tamtamresearch.com

            # alias name must not contain "/"
            [aliases]
            plcsr: %(pl-base)s/region/pl/ConsumerServiceReady.xml
            czcsr: %(cz-base)s/region/cz/ConsumerServiceReady.xml

        The format follows SafeConfigParser rules, see
        http://docs.python.org/2/library/configparser.html#safeconfigparser-objects

        To list all versions of czcsr alias::

            $ s3lsvers czcsr
    """
    cmdname = os.path.basename(sys.argv[0])

    bucket_name, key_name = parse_bucket_key(bucket_key)

    from_time, to_time = check_from_to(from_time, to_time)

    session = boto3.session.Session(profile_name=profile_name,
                                    aws_access_key_id=aws_access_key_id,
                                    aws_secret_access_key=aws_secret_access_key)
    s3 = session.resource("s3")
    bucket = s3.Bucket(bucket_name)

    # ver_filter = {"KeyMarker": key_name, "Prefix": key_name}
    ver_filter = {"Prefix": key_name}
    if version_id:
        ver_filter["KeyMarker"] = key_name
        ver_filter["VersionIdMarker"] = version_id

    versions = bucket.object_versions.filter(**ver_filter)

    try:
        json_data = []
        csvwriter = get_csvwriter(list_file)

        for ver in vergen(versions, key_name, from_time, to_time, working, 600):
            msg = "Date: {last_modified} Size: {size} Age: {age}"
            print(msg.format(**ver))
            if csvwriter:
                csvwriter.writerow(ver)
            if html_file:
                json_data.append(get_json_record(ver[
                                 "atime"], ver["size"], ver["age"]))

        if html_file:
            title = "{bucket_name}:{key_name}".format(**locals())
            json_data = json.dumps(json_data, separators=(",", ":"))
            res = render_html(json_data, title=title)
            html_file.write(res)
    except KeyboardInterrupt:
        print("...cancelled.")
    finally:
        if list_file and (not list_file.closed):
            list_file.close()
        if html_file and (not html_file.closed):
            html_file.close()
    return


def working(ver):
    msg = "version num: {i}, modified: {last_modified}"
    print(msg.format(**ver))


def vergen(versions, key_name, from_time, to_time,
           callback=None, callback_after=600):
    """return data
    atime, ka.size. age
    ka.name, ka.version_id, ka.size, ka.last_modified, age

    """
    vers_a, vers_b = itertools.tee(versions)
    next(vers_b)

    def buildver():
        """create dict, describing one version of key"""
        fmt = "{:%Y-%m-%dT%H:%M:%S}.000Z"
        return {"i": i, "atime": atime, "btime": btime, "age": age,
                "key_name": ka.key, "size": ka.size,
                "last_modified": fmt.format(ka.last_modified),
                "version_id": ka.version_id
                }
    for i, ka, kb in zip(itertools.count(), vers_a, vers_b):
        # if there are keys with the same prefix, we must stop
        # as soon as key_name does not match exactly
        if ka.key != key_name:
            break
        atime = ka.last_modified
        # if there are keys with the same prefix, we report age for last
        # key as 0
        last_modified = atime
        if kb.key == key_name:
            btime = kb.last_modified
            age = (atime - btime).seconds
        else:
            btime = None
            age = 0
        if from_time and (last_modified < from_time):
            break
        elif to_time and (last_modified > to_time):
            if i % callback_after == 0:
                callback(buildver())
            continue
        yield buildver()


def render_html(
        json_data,
        title,
        subtitle="(gzipped) feed size [bytes] and update intervals[seconds]",
        set_1_name="Feed size",
        set_2_name="Update iterval",
        ):
    return templ.format(**locals())


def placer():
    try:
        plac.call(main)
    except Exception as e:
        print(e)
        raise


if __name__ == "__main__":
    placer()



import pytest
from datetime import datetime
import pytz
from ttr.aws.utils.s3.utils import check_from_to


def utc(*args, **kwargs):
    res = datetime(*args, **kwargs)
    return res.replace(tzinfo=pytz.UTC)


scenarios = [
    ["2015-08-15T00:00:00Z", "2015-09-15T00:00:00Z",
     (utc(2015, 8, 15, 0, 0, 0),
      utc(2015, 9, 15, 0, 0, 0))],
    ["", "", (None, None)],
    ["", "2015-09-15T00:00:00Z", (None, utc(2015, 9, 15, 0, 0, 0))],
    ["2015-08-15T00:00:00Z", "", (utc(2015, 8, 15, 0, 0, 0),
                                  None)],
    pytest.param(
        ["2015-09-15T00:00:00Z", "2015-08-15T00:00:00Z",
         (utc(2015, 8, 15, 0, 0, 0),
          utc(2015, 9, 15, 0, 0, 0))],
        marks=pytest.mark.xfail), ]


def scen_id(scenario):
    from_txt = scenario[0]
    to_txt = scenario[1]
    return "{from_txt}-{to_txt}".format(from_txt=from_txt,
                                        to_txt=to_txt)


@pytest.mark.parametrize("scenario", scenarios, ids=scen_id)
def test_from_to(scenario):
    from_time, to_time, expected = scenario
    res = check_from_to(from_time, to_time)
    assert res == expected



import sys
from datetime import datetime

import plac
import boto3.session
import botocore.exceptions


# just to keep this code python 2.6 compatible
def total_seconds(timedelta):
    return (
        timedelta.microseconds + 0.0 +
        (timedelta.seconds + timedelta.days * 24 * 3600) * 10 ** 6) / 10 ** 6


def error(*objs):
    print("ERROR", *objs, end='\n', file=sys.stderr)


@plac.annotations(
    profile_name=("""Name of AWSCLI profile to use for credentials""",
                  "option"),
    aws_access_key_id=("Your AWS Access Key ID", "option"),
    aws_secret_access_key=("Your AWS Secret Access Key", "option"),
    validate_bucket=("Make sure, the bucket really exists", "flag"),
    validate_key=("Make sure, the key really exists", "flag"),
    http=("Force the url to use http and not https", "flag"),
    expire_dt=("ISO formatted time of expiration, full seconds,"
               " 'Z' is obligatory, e.g. '2014-02-14T21:47:16Z'"),
    bucket_name="name of bucket", key_names="key names to generate tmpurl for")
def main(
        expire_dt,
        bucket_name,
        profile_name=None,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        validate_bucket=False,
        validate_key=False,
        http=False,
        *key_names):
    """Generate temporary url for accessing content of AWS S3 key.

    Temporary url includes expiration time, after which it rejects serving the
    content.

    Urls are printed one per line to stdout.

    For missing key names empty line is printed and error goes to stderr.

    If the bucket is versioned, tmp url will serve the latest version
    at the moment of request (version_id is not part of generated url).

    By default, bucket and key name existnence is not verified.

    Url is using https, unless `-http` is used.
    """
    session = boto3.session.Session(profile_name=profile_name,
                                    aws_access_key_id=aws_access_key_id,
                                    aws_secret_access_key=aws_secret_access_key)
    s3client = session.client("s3")
    s3 = session.resource("s3")
    expire_dt = datetime.strptime(expire_dt, "%Y-%m-%dT%H:%M:%SZ")

    params = {"Bucket": bucket_name}
    if http:
        params["HttpMethod"] = "http"

    if validate_bucket or validate_key:
        try:
            bucket = s3.Bucket(bucket_name)
            bucket.creation_date
        except botocore.exceptions.ClientError:
            error("Error: Bucket not found: ", bucket_name)
            return
    for key_name in key_names:
        if validate_key:
            try:
                key = s3.Object(bucket_name, key_name)
                key.get()
            except botocore.exceptions.NoSuchKey:
                error("Error: missing key: ", key_name)
                print("")
                continue
        # get a key object without version_id
        # (otherwise we get url to specific version)
        params["Key"] = key_name
        url = s3client.generate_presigned_url(
            "get_object",
            Params=params,
            ExpiresIn=total_seconds(expire_dt-datetime.utcnow()))
        print(url)


def placer():
    try:
        plac.call(main)
    except Exception as e:
        error(e)

if __name__ == "__main__":
    placer()



# import dateutil.parser
import pendulum
from ttr.aws.utils.s3.saver import FnameFactory
import pytest


def get_last_modified_scenarios():
    plan = [["buck/k/n.json:1234aaa:2017-05-15T20:25:07Z",
             "n.2017-05-15T20_25_07Z.json"],
            ["buck/d.json:1234aaa:1970-01-01T00:00:00Z",
             "d.1970-01-01T00_00_00Z.json"],
            ]
    for objver, expected in plan:
        obj, version_id, dtstr = objver.split(":", 2)
        bucket_name, key_name = obj.split("/", 1)
        dt = pendulum.parse(dtstr)
        objverdct = {"LastModified": dt, "VersionId": version_id}
        yield bucket_name, key_name, version_id, objverdct, expected


@pytest.mark.parametrize("scenario", list(get_last_modified_scenarios()))
def test_last_modified(scenario):
    bucket_name, key_name, version_id, objdict, expected = scenario
    fname_factory = FnameFactory().last_modified
    fname = fname_factory(bucket_name, key_name, version_id, objdict)
    assert fname == expected


def get_version_id_scenarios():
    plan = [["buck/k/n.json:1234aaa:2017-05-15T20:25:07Z",
             "n.1234aaa.json"],
            ["buck/d.json:1234aaa:1970-01-01T00:00:00Z",
             "d.1234aaa.json"],
            ]
    for objver, expected in plan:
        obj, version_id, dtstr = objver.split(":", 2)
        bucket_name, key_name = obj.split("/", 1)
        dt = pendulum.parse(dtstr)
        objverdct = {"LastModified": dt, "VersionId": version_id}
        yield bucket_name, key_name, version_id, objverdct, expected


@pytest.mark.parametrize("scenario", list(get_version_id_scenarios()))
def test_version_id(scenario):
    bucket_name, key_name, version_id, objdict, expected = scenario
    fname_factory = FnameFactory().version_id
    fname = fname_factory(bucket_name, key_name, version_id, objdict)
    assert fname == expected



import time
import os.path
from .utils import gunzip_data


class ObjectVersionSaver(object):
    """Class for saving S3 object versions on demand.

    >>> fname_factory = FnameFactory().last_modified
    >>> saver = ObjectVersionSaver(s3, fname_factory, to_decompress)
    >>> fname, keydct = saver.save("buck", "a/b.json", "da13b")
    >>> fname, keydct = saver.save("buck", "a/b.json", "ba26c")
    >>> fname, keydct = saver.save("buck", "a/c.json", "dd112")

    fname is file name where given version was saved.
    keydct is dict as from s3.ObjectVersion(...).get()
    """
    def __init__(self, s3, fname_factory, to_decompress):
        """
        Args:
            s3 (boto3.resources.factory.s3.ServiceResource):
                >>> from boto3.session import Session
                >>> session = Session()  # use access keys, profile_name etc
                >>> s3 = session.resource("s3")
            fname_factory (func returning file name): The function takes args:
                bucket_name (str): S3 bucket name
                key_name (str): S3 object key name
                version_id (str): S3 object version identifier
                objdict (dict): dict from s3.ObjectVersion(...).get()
            to_decompress (tuple of content encodings): content encodings
                to decompress. Usually None or ("deflate", "gzip").
                Only "deflate" and "gzip" encodings supported at the moment.
        """
        self.s3 = s3
        self.fname_factory = fname_factory
        self.to_decompress = to_decompress

    def save(self, bucket_name, key_name, version_id):
        """
        Save given object version into file (created by self.fname_factory).
        Args:
            bucket_name (str): S3 bucket name
            key_name (str): S3 object key name
            version_id (str): S3 object version identifier
        Raises:
            botocore.exceptions.ClientError: when missing credentials, missing
                bucket, missing object or object version or insufficient
                permissions.
        Returns:
            tuple (fname, keydct):
                fname (str): file name
                keydct (dict): as returned from s3.ObjectVersion(...).get()
        """
        key = self.s3.ObjectVersion(bucket_name, key_name, version_id).get()
        last_modified = key["LastModified"]
        fname = self.fname_factory(bucket_name,
                                   key_name,
                                   version_id,
                                   key)
        # TODO: modify for large files into streaming
        # gzip requires .tell and key["Body"] does not provide it
        with open(fname, "wb") as f:
            if key["ContentEncoding"] in self.to_decompress:
                f.write(gunzip_data(key["Body"].read()))
            else:
                f.write(key["Body"].read())
        # set modification time to the time, version was created in bucket
        tm = time.mktime(last_modified.timetuple())
        os.utime(fname, (tm, tm))
        return fname, key


class FnameFactory(object):
    """Provide functions creating for S3 object version a file name.
    Such filename is usually used to save the content into it.

    Currently it provides two different functions:
        last_modified: name contains (not only) time of last modification.
        version_id: name contains (not only) object version_id.
    """

    def __init__(self):
        pass

    def last_modified(self, bucket_name, key_name, version_id, objdict):
        """Create for S3 object version unique file name.

        e.g.: "data.2017-05-15T22_21_07Z.json"

        Args:
            bucket_name (str): S3 bucket name
            key_name (str): S3 object key name
            version_id (str): S3 object version identifier
            objdict (dict): dict from s3.ObjectVersion(...).get()
        Returns (str): file name
        """
        basename, ext = os.path.splitext(os.path.basename(key_name))
        templ = "{basename}.{LastModified:%Y-%m-%dT%H_%M_%SZ}{ext}"
        return templ.format(basename=basename,
                            ext=ext,
                            **objdict)

    def version_id(self, bucket_name, key_name, version_id, objdict):
        """Create for S3 object version unique file name.

        e.g.: "data.d4b4cb144387dqf.json"

        Args:
            bucket_name (str): S3 bucket name
            key_name (str): S3 object key name
            version_id (str): S3 object version identifier
            objdict (dict): dict from s3.ObjectVersion(...).get()
        Returns (str): file name
        """
        basename, ext = os.path.splitext(os.path.basename(key_name))
        templ = "{basename}.{version_id}{ext}"
        return templ.format(basename=basename,
                            ext=ext,
                            version_id=version_id,
                            **objdict)



import gzip
import datetime
import pytz
from io import BytesIO


def gunzip_data(data):
    '''Decomress data from string using gzip method.

    :type data: string
    :param data: string to be decompressed

    :rtype: string
    :return: decompressed string
    '''
    buf = BytesIO(data)
    gzipper = gzip.GzipFile(fileobj=buf)
    data = gzipper.read()
    gzipper.close()
    # todo, check for possible memory leak, shall I close the buf?
    return data


def check_from_to(from_txt, to_txt):
    if from_txt:
        from_time = parse_datetime(from_txt, "min")
    else:
        from_time = None

    if to_txt:
        to_time = parse_datetime(to_txt, "max")
    else:
        to_time = None

    if (from_time and to_time):
        if not from_time < to_time:
            msg = "from ({from_txt}) must be smaller then -to ({to_txt})"
            raise ValueError(msg.format(from_txt=from_txt, to_txt=to_txt))
    return from_time, to_time


def parse_datetime(text, padmode):
    padmodes = ["min", "max"]
    if padmode not in padmodes:
        raise ValueError('padmode must be `"min"` or `"max"`')
    padtxts = {"min": ["0000-01-01T00:00:00Z"],
               "max": ["9999-99-31T23:59:59Z",
                       "9999-99-30T23:59:59Z",
                       "9999-99-29T23:59:59Z",
                       "9999-99-28T23:59:59Z",

                       "9999-12-31T23:59:59Z",
                       "9999-12-30T23:59:59Z",
                       "9999-12-29T23:59:59Z",
                       "9999-12-28T23:59:59Z"
                       ]
               }[padmode]
    fmt = "%Y-%m-%dT%H:%M:%S"
    exc = None
    for padtxt in padtxts:
        ftext = text + padtxt[len(text):]
        try:
            res = datetime.datetime.strptime(ftext[:19], fmt)
            res = res.replace(tzinfo=pytz.UTC)
            return res
        except ValueError as e:
            # expect e.message "day is out of range for month"
            exc = e
            pass
    if exc is not None:
        raise exc
